"""
Created the 2021/10/08
v0.0 First version
v0.1 Update graph by line instead of scatter
v0.2 add multirenaming option
v0.3 naming fix from plotyfermy compatibility
v0.4 add time slice option to fanodygrowth function
v0.5 add method to average replicates
v0.6 bug fix for replicates
v0.7 add EGRA % max slope option / add max OD to fermgrowth
v0.8 replace groupeby axis = 1 per .T
==> next step add minimal DeltaBiomassProxy for fanodygrowth function
==> next step calcul max Proxxy at max slope
    if dfslope.min()<limite:
            indexODmaxslope = dfslope.loc[dataafterslope<=0.005].first_valid_index()
==> next step: add slope calculation method as pandas method

@author: Nicolas Hardy

This file is part of Fermy.

    Fermy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Fermy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Fermy.  If not, see <https://www.gnu.org/licenses/>.
"""

__version__ = 0.8


import pandas as pd
import numpy as np
from typing import List, Dict, Iterable, Tuple, Union
import os
import datetime
import argparse

from fermy.log import logger

def linearareaselection(data: pd.Series, percent: float=0.95) -> pd.Series:
    """"Function to return series that map best lineare area
    return the table wiht na or slope. slopes area where it is linear
    """
    maxslope = data.max() * percent
    return data.where(data>maxslope).dropna()

def EGRAdataselection(data: pd.Series, originaldataset: pd.DataFrame) -> pd.Series:
    """Function to return series that map with EGRA
    return orginialdataset filter by non na data values
    """
    EGRA = originaldataset[data.name].loc[data.dropna().index]  #select data from smooth data matching with EGRA
    return EGRA

def normalizedata(data: pd.Series, regardingmax:bool=True, percentofmax:float=0.05, numberofpointformin:int=5) -> pd.Series:
    """Function to normilized by the average value of first five points not null
    or if regardingmax = True: normilized by percentofmax of maximum value
    if data is low than normalized value put it at 1
    """
    if regardingmax:
        logger.info("By using max for normalization\n growth rate estimation will be more robust\nwhereas lag time will no more have meaning")
        #print("By using max for normalization\n growth rate estimation will be more robust\nwhereas lag time will no more have meaning")
        normalvalue = data.max() * percentofmax
        normalizeddata = data/normalvalue
    else:
        normalvalue = data.replace(to_replace=0, value=np.nan).dropna().iloc[:numberofpointformin].mean()
        normalizeddata = data/normalvalue
    normalizeddata = [1 if valeur < 1 else valeur for valeur in normalizeddata]  # if value <1 replace it by 1
    return normalizeddata

def reglin(valuesx: np.ndarray, valuesy: np.ndarray, slopeonly=False) -> Union[float, List[float]]:
    """Minimal linear regression with numpy. You can set slopeonly at True
    to return only slope else return both
    """
    slope, intercept = np.polyfit(valuesx, valuesy, 1)  # deg 1: ax+b
    if slopeonly:
        return slope
    else:
        return [float(slope), float(intercept)]

def calcslope(data: pd.Series, slopeonly=True) -> Union[float, List[float]]:
    """Function to compute slope from a Serie
    return slope or list with slope and intercept
    """
    data.dropna(inplace=True)
    slope = reglin(data.index.values, data.values, slopeonly=slopeonly)
    return slope

def localmax(data: pd.DataFrame) -> pd.Series:
    """Function to found local max of slopes
    """
    peak_df = data[(data.shift(1) < data) & (data.shift(-1) < data)]
    return peak_df


def multireg(data: pd.Series, windows: float=5)-> List[float]:
    """Function to list mu (slope) and lagtime (-intercepts/slope)
    return list of mu, list of lagtime and list of time to mu
    need series and index of slopes
    """
    indexlocalmax = (
            data
            .rolling(5, center=True).apply(calcslope)  # compute rolling slope
            .rolling(windows, center=True).mean()  # smooth to reduce local maximum number
            .to_frame()  # trick because local comput only on DF
            .apply(localmax)
            .dropna().index # Local max index
            )
    listmu = []
    lagtime = None
    firstlagtimeflag = True  #to know if it is the first lagtime
    for slopeindex in indexlocalmax:
        indexlist = []
        slopeindexnum = data.index.get_loc(slopeindex)
        for switchtnum in range(-2,3):
            indexlist.append(slopeindexnum+switchtnum)
        x, y = data.iloc[indexlist].index, data.iloc[indexlist].values
        listmu.append(round(reglin(x,y)[0],2))
        if firstlagtimeflag:
            lagtime = round(-reglin(x,y)[1]/reglin(x,y)[0],2)
            if lagtime:
                firstlagtimeflag = False  # Check if lagtime exist
    return [listmu, lagtime, [round(time, 2) for time in indexlocalmax.to_list()]]



def selectdatawithminidelta(data: pd.Series, deltathreshold: float) -> pd.Series:
    """"Function to return DataFrame that have at least deltathreshold
    by comparing min and max in column if not or an empty dataseries
    """
    if data.max()-data.min() > deltathreshold:
        return data
    else:
        return pd.Series(data=np.nan, index=data.index, name=data.name)



def fanodygrowth(data: pd.DataFrame, percentofmax:float=0.05, usemax:bool=False, timethreshold:Union[None, float]=None, deltathreshold:Union[None, float]=None , percentegra:float=0.95) -> pd.DataFrame:
    """Function to compute growth rate and lagtime
    It is based on a mix of two algorithms describes in the following:
    Toussaint et al. 2006 and Hall et al. 2014.
    
    Input:
    -----
    A DataFrame with only data considered as proxy of biomass can be used.
    DataFrame index has to be a float corresponding to
    a delta time (in hours) or pandas.DatetimeIndex
    
    Parameters:
    ----------
    usemax: allow users to choose between normalization methods:
            average of the first five points
            or 
            percentage of the maximal value of dataset
    
    percentofmax: allow users to change the percentage of
            maximal value used
    
    timethreshold: allow users to compute growth rate without
            considering data from time before
            the provided number of hours
    
    deltathreshold: allow users to exclude data without a minimal
                Biomass proxy delta between min and max of the column.
    
    percentegra: allow users to change the % used to select the
                relevent slope area (EGRA) variation of max slope
                by default 95% (0.95) of max slope
    
    Output:
    ------
    return a DataFrame with lagtime and growth rate
    """
    datatemp = data.copy()
    if isinstance(datatemp.index, pd.DatetimeIndex):
        # set index as derltatime in hours
        datatemp.index = datatemp.index-datatemp.index[0]  # cal detlatime
        datatemp.index = datatemp.index.total_seconds()/3600 #index in hour in decimal format
    elif datatemp.index.dtype == float:
        logger.info("We assumed a DataFrame with float in hours as index")
        #print("We assumed a DataFrame with float in hours as index")
    else:
        logger.info("We assumed a DataFrame with float in hours as index")
        #print("Please provide a Dataframe with Datetime as index\n")
    
    if timethreshold:  # option to not use data before timethreshold time
        datatemp = datatemp.loc[datatemp.index >= timethreshold].copy()
        logger.info(f"Data for time before {timethreshold}h does not use for calculation due to user request, so lagtime maybe be wrong.")
        #print(f"Data for time before {timethreshold}h does not use for calculation due to user request, so lagtime maybe be wrong.")
    
    if deltathreshold:
        # Discare data without a minimal delta max - min in column
        datatemp = datatemp.apply(selectdatawithminidelta, args=[deltathreshold]).dropna(axis=1, how="all")
    
    #Legacy algo from table data with decimal deltatime in index and gowth proxy in columns help to Pandas pipe
    #clean data set
    datasmooth = (
        datatemp
        .apply(normalizedata, args=(usemax, percentofmax))  # normalized data by first five point average
        .apply(np.log)  # ln(data/datamin)
        .rolling(window=9, center=True).mean()  # smooth ln(data) window of 9
    )
    
    #found index of slope max
    indexmax = (
    datasmooth
    .rolling(5, center=True).apply(calcslope)  # compute rolling slope
    .idxmax()
    )
    
    #Compute growthrate and lagtime
    fanody = (
        datasmooth
        .rolling(5, center=True).apply(calcslope)  # compute rolling slope
        .apply(linearareaselection, percent=percentegra) # selection of relevent slope area as % variation of max slope EGRA
        .apply(EGRAdataselection, args=(datasmooth,))  # selection of relevent data from smotth table to compute µ
        .apply(calcslope, args=(False,)).T  # compute Slope and Intercept and transposed
        .rename(columns = {0:"maximal_growth_rate_per_h", 1 : "lagtime_h"})  # renaming
        .assign(lagtime_h = lambda df: -df["lagtime_h"]/df["maximal_growth_rate_per_h"])  #compute lagtime (basted on intercept already and wrongly named lagtime)
        .assign(maximal_growth_rate_time_h = indexmax)  #add time of mumax
        .assign(maximal_od = datatemp.max())  # max od of original dataframe
        .round(2)  #round at two decimals
            )
    
    return fanody


def fanodymultiauxic(data:pd.DataFrame, windows:float=5, percentofmax: float=0.05, usemax:bool=False) -> pd.DataFrame:
    """Function to compute growths rates and lagtimes for multiauxies
    It is based on a mix of two algorithms describes in the following:
    Toussaint et al. 2006 and Hall et al. 2014.
    Whereas growth rates are found with local max computation.
    
    Input:
    -----
    A DataFrame with only data considered as proxy of biomass can be used.
    DataFrame index has to be a float corresponding to
    a delta time (in hours) or pandas.DatetimeIndex
    
    Parameters:
    ----------
    usemax: allow users by use chose between normalization methods:
            average of the first five points
            or 
            percentage of the maximal value of dataset
    percentofmax: allow users to change the percentage of maximal value used
    
    Output:
    ------
    return a DataFrame with lagtimes and growth rates
    """
    datatemp = data.copy()
    if isinstance(datatemp.index, pd.DatetimeIndex):
        # set index as derltatime in hours
        datatemp.index = datatemp.index-datatemp.index[0]  # cal detlatime
        datatemp.index = datatemp.index.total_seconds()/3600 #index in hour in decimal format
    elif datatemp.index.dtype == float:
        logger.info("We assumed a DataFrame with float in hours as index")
        #print("We assumed a DataFrame with float in hours as index")
    else:
        logger.info("Please provide a Dataframe with Datetime as index\n")
        #print("Please provide a Dataframe with Datetime as index\n")
    #FanODy algo from table data with decimal deltatime in index and gowth proxy in columns help to Pandas pipe
    #clean data set
    datasmooth = (
        datatemp
        .apply(normalizedata, args=(usemax, percentofmax))  # normalized data by first five point average
        .apply(np.log)  # ln(data/datamin)
        .rolling(window=9, center=True).mean()  # smooth ln(data) window of 9
    )
    
    #compute growthrate and lagtimes for multiauxic growth
    fanodymulti =(
                datasmooth.apply(multireg, args=(windows,)).T # compute Slopes and Intercepts and transposed
                .rename(columns = {0:"maximal_growth_rate_per_h", 1 : "lagtime_h", 2 : "maximal_growth_rate_time_h"})  # renaming
                )
                
                
    return fanodymulti


def builddicorename(DFdata : pd.DataFrame, dicorenamingrullsuserdef: Dict) -> Dict:
    """Function to build a specific rename dicotionary for DataFrame columns names
    input: DFdata a Pandas DataFrame
            dicorull with key as new column name and values differents values possibly present in
            the dataframe
    output: the dictionary to achieve dataframe.rename(columns=builddicorename(dataframe, dicorenamingrullsuserdef), inplace=True)
    """
    columnlist = DFdata.columns.tolist()
    listvalues = list(dicorenamingrullsuserdef.values())
    listkeys = list(dicorenamingrullsuserdef.keys())
    dicorename = {}
    for columname in columnlist:
        for index, values in enumerate(listvalues):
            if columname in values:
                dicorename[columname] = listkeys[index]
    return dicorename

def customedrenamefunction(DFdata:pd.DataFrame, dicorenamingrullsuserdef:Dict, inplace:bool=False) -> pd.DataFrame:
    """Function to rename from a dictionary in forme {"newname" : (possibleoldname1, possibleoldname2)}
    and rename columns in a DataFrame like possiblename1 ==> new name.
    
    Input:
    -----
    DataFrame
    Dictionary with key as new column name and
    values list of values possibly used in the Dataframe as columns
    
    Output:
    ------
    return a DataFrame with new columns names
    """
    outDF = DFdata.rename(columns=builddicorename(DFdata, dicorenamingrullsuserdef), inplace=inplace)
    return outDF

def meanreplicates(dataframe:pd.DataFrame, sepchar:str):
    """Function to average data from a pandas DataFrame grouping by left part of column name
    after splitting by a defined separated character (sepchar)
    Input:
    -----
    DataFrame
    The character that separate base name to sample identification
    
    Output:
    ------
    return a DataFrame with averaged data per base name
    
    """
    listofsamplebasename = [column.split(sepchar)[0] for column in dataframe.columns]
    listofsamplename = dataframe.columns.to_list()
    if len(listofsamplebasename) == len(listofsamplename):
        mapping = dict(zip(listofsamplename, listofsamplebasename))
        return dataframe.T.groupby(by=mapping).mean().T
    else:
        return dataframe


pd.core.base.PandasObject.fermgrowth = fanodygrowth # monkey-patch the DataFrame class to add growth rate calculation
pd.core.base.PandasObject.multirename = customedrenamefunction # monkey-patch the DataFrame class to add multiranaming function
pd.core.base.PandasObject.fermmultiaux = fanodymultiauxic # monkey-patch the DataFrame class to add multiphasic growth rates calculation
pd.core.base.PandasObject.replicates = meanreplicates # monkey-patch the DataFrame class to add capacity to average by replicates identified by column names


if __name__ == '__main__':
    """here argparser code"""
    parser = argparse.ArgumentParser(description = "Pandasfermy add to Pandas method to compute analysis on fermentation data", 
    epilog="For Help or more information please contact Nicolas Hardy")
    
    parser.add_argument("filepath", metavar = "Root of data", type = str, help = "File with fermentation data first column have to be time")
    parser.add_argument("-p","--percent", dest="percentofmax", type = float, default = 0.05, help = "Percentage of the maximal value considered as relevant by default it is egal to 0.05 (for 5 percent)")
    parser.add_argument("-m","--min", action='store_false', dest="usemax", help = "Option to use mininaml value for nomalization (like legacy algorithm ) instead of maximal value percentage")
    
    args = parser.parse_args()
    
    percentofmax = args.percentofmax
    filepath = args.filepath
    usemax = args.usemax
    """"code here"""
    #load data
    data = pd.read_excel(filepath, index_col=0)  #load data with Pandas index have to be datetime.datetime
    #use the pandasfermy
    print(usemax)
    print(f"percent of max value for nomalization will be {percentofmax*100}%")
    fanody = data.fermgrowth(percentofmax=percentofmax ,usemax=usemax)
    multiaux = data.fermmultiaux(windows=5,percentofmax=percentofmax, usemax=usemax)
    print(fanody, multiaux)
    #Save data
    fanody.to_excel(os.path.splitext(filepath)[0]+f"{percentofmax}_out.xlsx")
    multiaux.to_excel(os.path.splitext(filepath)[0]+f"{percentofmax}_multiaux_out.xlsx")



"""
Created the 2022/10/04
v0.1 First version
v0.2 Bugfix with Epoch2 and fakedata name
v0.3 add Tecan reader 96 and 384
v0.4 add Logphase600 96 BioTek & add exclusion of rows with nan
v0.5 bug fix for method .applymap to .map in fuction readepochtwo ==> need Pandas > 2.1
v0.6 optimize readepochtwo to skip the end of the file by finding result instead of fixed value 
v0.7 add Fluostar Omega or Optima / add bug fix for Tecan when column missing
v0.8 add index name as "Time" for Fluostar and minor fix for convfluostartohours
v0.9 add maptoserie function
v1.0 add maptoserie with zero as method
@author: Nicolas Hardy



This file is part of Fermy.

    Fermy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Fermy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Fermy.  If not, see <https://www.gnu.org/licenses/>.
"""

__version__ = 1.0

import pandas as pd
from typing import List, Dict, Iterable, Tuple, Union
import math

class MTPError(Exception):
    def __init__(self, message):
        super().__init__(message)



class MTPReader():
    """Class that allow to read MTP data need filepath and readername
    """
    
    def __init__(self, filepath:str, readername:str):
        """Initialisation of reader
        """
        self.readername = readername
        self.plate = pd.DataFrame()
        self.filepath = filepath
        self.dicoreader = {"Epoch2" : self.readepochtwo,
                            "fakeMTP" : self.readfakemtp,
                            "Tecan96" : self.readtecan96,
                            "Tecan384" : self.readtecan384,
                            "Logphase" : self.readlogphasesixhundr,
                            "Fluostar" : self.readfluostar
                            } # linker between MTP and reading function
    
    def convstringtohours(self, string: str) -> float:
        """"Function to transform a string like hours:minutes:seconds
        or aaa-mm-dd hh:mm:ss to a float of hours
        return float
        """
        if len(string.split(" ")) == 1:
            spliting = string.split(':')
            timefloat = float(spliting[0])+float(spliting[1])/60+float(spliting[2])/3600
        else:
            spliting = string.split(" ")
            timefloat = float(spliting[0].split('-')[2])*24
            timefloat = timefloat + float(spliting[1].split(':')[0])
            timefloat = timefloat + float(spliting[1].split(':')[1])/60
            timefloat = timefloat + float(spliting[1].split(':')[2])/3600
        return timefloat
    
    def convfluostartohours(self, string: str) -> float:
        """"Function to transform a string like hours:minutes
        or H h M min to a float of hours
        return float
        """
        timefloat = 0.
        if "h" in string:
            spliting = string.split("h")
            for data in spliting:
                if "min" in data:
                    timefloat += float(data.strip(" min"))/60
                elif data.strip(" ") != "":
                    timefloat += float(data.strip(" "))
        if ":" in string:
            spliting = string.split(":")
            if len(spliting) == 3:
                timefloat += float(spliting[0])
                timefloat += float(spliting[1])/60
                timefloat += float(spliting[2])/3600
        else:
            timefloat += 0.
        return timefloat
    
    def buildrenamer(self, inputdata:Union[None,List] = None,switch:int=0) -> Dict:
        """Function to build a MTP 96 well renamer
        return dico
        """
        dicorenamer = {}
        if inputdata != None:
            index = 0
            for line in ["A", "B", "C","D", "E", "F", "G", "H"]:
                for column in range(1,13):
                    dicorenamer[inputdata[index]] = f"{line}{column}"
                    index +=1
            return dicorenamer
        else:
            index = 0 + switch
            for line in ["A", "B", "C","D", "E", "F", "G", "H"]:
                for column in range(1,13):
                    dicorenamer[index] = f"{line}{column}"
                    index +=1
            return dicorenamer
    
    
    def readepochtwo(self) -> pd.DataFrame:
        """Function to return formated dataframe from reader Epoch
        index is time float in hours
        """
        
        rawtable = pd.read_excel(self.filepath, sheet_name=0)
        numline = int(rawtable[rawtable["Unnamed: 0"]=="Layout"].index.values) +2
        numfin = int(rawtable[rawtable["Unnamed: 0"]=="Results"].index.values)
        filefin = len(rawtable.index)+1
        endskipe = filefin-numfin
        
        del rawtable
        tabledata = pd.read_excel(self.filepath, sheet_name=0, index_col=None, skiprows=numline+13, skipfooter=endskipe, dtype={"Time":str}).iloc[:,1:]  #data robot
        tabledata["Time"] = tabledata["Time"].apply(self.convstringtohours)  # converrt time from hh:mm:ss to float
        tabledata.index = tabledata["Time"] - tabledata["Time"].iloc[0] # set index start from time 0.0 h
        tabledata.drop(["Time", "T° 600"], axis=1, inplace=True, errors = "ignore")  # Drop Time columns
        tabledata = tabledata.map(lambda data: data.strip("*") if isinstance(data, str) else data)  # clean *
        tabledata = tabledata.astype('float64')
        tabledata.dropna(axis=0, how="all", subset=tabledata.columns[2:], inplace=True)  # clear empty data
        tabledata.index.name = "Time"
        return tabledata
    
    def readtecan96(self) -> pd.DataFrame:
        """Function to return formated dataframe from reader Tecan with 96
        index is time float in hours
        """
        rawtable = pd.read_excel(self.filepath, sheet_name=0)
        numlinedata = int(rawtable[rawtable.iloc[:,0] == "Cycle Nr."].index.values)
        numlineenddata = int(rawtable[rawtable.iloc[:,0] == "End Time"].index.values[-1])
        numlineendtoskip = rawtable.index.max() - numlineenddata + 2
        del rawtable
        tabledata = pd.read_excel(self.filepath, sheet_name=0, index_col=None, skiprows=numlinedata+1, skipfooter=numlineendtoskip).iloc[:,1:]  #data plate 96
        #tabledata = pd.read_excel(self.filepath, sheet_name=0, usecols="B:CU", index_col=None, skiprows=numlinedata+1, skipfooter=numlineendtoskip)  #data plate 96
        tabledata["Time"] = tabledata["Time [s]"]/3600 # conversion to hours
        tabledata.index = tabledata["Time"]  # set time as index
        tabledata.drop(["Time", "Temp. [°C]", "Time [s]"], axis=1, inplace=True, errors = "ignore")  # Drop Time columns
        return tabledata
    
    def readtecan384(self) -> pd.DataFrame:
        """Function to return formated dataframe from reader Tecan with 96
        index is time float in hours
        """
        rawtable = pd.read_excel(self.filepath, sheet_name=0)
        numlinedata = int(rawtable[rawtable.iloc[:,0] == "Cycle Nr."].index.values)
        numlineenddata = int(rawtable[rawtable.iloc[:,0] == "End Time"].index.values[-1])
        numlineendtoskip = rawtable.index.max() - numlineenddata + 2
        del rawtable
        tabledata = pd.read_excel(self.filepath, sheet_name=0, index_col=None, skiprows=numlinedata+1, skipfooter=numlineendtoskip).iloc[:,1:]  #data plate 384
        #tabledata = pd.read_excel(self.filepath, sheet_name=0, usecols="B:NW", index_col=None, skiprows=numlinedata+1, skipfooter=numlineendtoskip)  #data plate 384
        tabledata["Time"] = tabledata["Time [s]"]/3600 # conversion to hours
        tabledata.index = tabledata["Time"]  # set time as index
        tabledata.drop(["Time", "Temp. [°C]", "Time [s]"], axis=1, inplace=True, errors = "ignore")  # Drop Time columns
        return tabledata
        
    def readlogphasesixhundr(self) -> pd.DataFrame:
        """Function to return formated dataframe from Logphase600
        from BioTek Agilent
        index is time float in hours
        """
        with open(self.filepath,"r") as filetemp:
            lines = filetemp.readlines()
            indexfile = 0
        for line in lines:
            if "Time\t" in line:
                numline = indexfile
                # print(f"{indexfile} in ''{line}'")
            indexfile += 1
        tabledata = pd.read_csv(self.filepath, skiprows=numline, encoding='cp1252', sep="\t", skip_blank_lines=True, decimal=",")
        tabledata.index = tabledata["Time"].apply(self.convstringtohours)  # converrt time from hh:mm:ss to float
        tabledata.drop(["Time"], axis=1, inplace=True, errors = "ignore")  # Drop Time columns
        return tabledata
        
    def readfakemtp(self) -> pd.DataFrame:
        """Function to return a fake formated dataframe
        """
        # time 5 minutes steps in hours for 6 hours
        time = [time/60 for time in range(0, 60*6, 5)]
        lagtime = time[20]  # 1.66 h
        fakedataset = [0.01]*20+[0.01*math.exp(0.5*(time-lagtime)) for time in time[20:]]
        datadico = {}
        for line in ["A", "B", "C","D", "E", "F", "G", "H"]:
            for column in range(1,13):
                datadico[f"{line}{column}"] = fakedataset
        tabledata = pd.DataFrame(data= datadico, index=time)
        tabledata.index.name = "Time"
        return tabledata
    
    def readfluostar(self) -> pd.DataFrame:
        """Function to return formated dataframe from Fluostar Optima
        or Omega
        index is time float in hours
        """
        rawtable = pd.read_excel(self.filepath, sheet_name=0)
        wellindex = rawtable[rawtable.iloc[:,0] == "Well"].index.to_list()
        if len(wellindex) == 1:
            numlinedata = int(wellindex[0])
            numcoldata = 2
            stringtimesep = ":"
        elif wellindex == []:
            wellindex = rawtable[rawtable.iloc[:,0] == "Well\nRow"].index.to_list()
            if len(wellindex) == 1:
                numlinedata = int(wellindex[0])
                numcoldata = 3
                stringtimesep = ["h","min"]
            else:
                raise MTPError(f"Corrupted file {self.filepath}")
        del rawtable
        tabledata = pd.read_excel(self.filepath, sheet_name=0, skiprows=numlinedata+2).iloc[:,numcoldata:].T
        tabledata.index = tabledata.index.to_series().apply(str).apply(self.convfluostartohours)
        tabledata.rename(columns=self.buildrenamer(tabledata.columns.to_list()), inplace=True)
        tabledata.index.name = "Time"
        return tabledata
    
    def readMTP(self) -> pd.DataFrame:
        """Load a file and provide a formated DataFrame
        """
        #read the input file if the reader name is known 
        if self.readername in self.dicoreader.keys():
            try:
                self.plate = self.dicoreader[self.readername]()
            except Exception as err:
                raise MTPError(f"Wrong device '{self.readername}' or corrupted file")
        self.plate.dropna(axis=0, how="any", inplace=True) # clean row with missing data (useless)
        return self.plate
    
    def readerslist(self) -> List[str]:
        """Provide the names of readers avaible
        """
        listofreaders = list(self.dicoreader.keys())
        return listofreaders
        
    def __str__(self):
        return f"The MTP reader is set as {self.readername}\nThe path of data file is {self.filepath}\nOther avaible readers {list(self.dicoreader.keys())}"


def maptoserie(dataframe: pd.DataFrame, forcezero=False) -> pd.Series:
    """Read plate map 96 wellplate per well
    from a dataframe with numbe 1-12 as column name and letter A:F as index
    In a simple Excel File: pd.read_excel(path,index_col=0,header=0)
    return a pd.serie with coord (e.g. A10) as index and name of well as value
    forcezero = True >> A1 => A01
    """
    newindex = []
    newvalue = []
    for raw, cols in dataframe.iterrows():
        for col in cols.index:
            if forcezero:
                if col <10:
                    newindex.append(f"{raw}0{col}")
                    newvalue.append(cols[col])
                else:
                    newindex.append(f"{raw}{col}")
                    newvalue.append(cols[col])
            else:
                newindex.append(f"{raw}{col}")
                newvalue.append(cols[col])
    labelserie = pd.Series(data=newvalue,index=newindex)
    return labelserie

pd.core.base.PandasObject.maptoserie = maptoserie # monkey-patch the DataFrame class to add maptoserie method



"""
Created the 2020/03/16
v0.1 First version
@author: Nicolas Hardy

This file is part of Fermy.

    Fermy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Fermy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Fermy.  If not, see <https://www.gnu.org/licenses/>.
"""
__version__ = 0.1

from labjack import ljm  # To solve it: python -m pip install labjack-ljm
import datetime
import math


class PompWatMarlo:
    """Class that connect and drive a Watson Marlow pump 120U plus head 114DV connect to
    a computer through a labjack T7
    connection DAC0:pin 2 GND:pin10 AIN0:pin9 GND:pin3
    link 4-20mA pin 1--11
    """
    def __init__(self):
        """Initialisation"""
        self.setpoint = 0
        self.rpm = 0
        self.connection = False
        self.info = None
        self.time = datetime.datetime.now()
        self.vdc_input = 0
        self.tensionread = 0
        #Calibration
        self.tachyA = 0.025
        self.tachyB = 0.0065
        self.rpmmax = 200
        self.rpmvdcA = 41.754
        self.rpmvdcB = -50.805
        self.limvdcsup = 3.75
        self.rpmvdcAsup = -27.73
        self.rpmvdcBsup = +253.84
        self.rpmvdcCsup = -456.79
    
    def __str__(self):
        """Define printing method print(object)"""
        return f"Pompe connection = {self.connection}\nlast read: {self.time} : {self.rpm}\nsetpoint {self.setpoint}"
        
    def polysecdeg_inv(self, y, a, b, c):
        """Quadratic function (y = a*x**2 + b*x + c ) solved with method of Completing the Squares
        Vertex = point of swapping domain Vertex(h,k)
        h from x
        k from y
        """
        s1 = -(b/(2*a))-math.sqrt(y/a-c/a+(b/(2*a))**2)
        s2 = (-b/(2*a))+math.sqrt(y/a-c/a+(b/(2*a))**2)
        h = -b/(2*a)
        k = c-a*(b/(2*a))**2
        if y < k:
            return s1
        elif y==k:
            return h
        else:
            return s2
    
    def rpmtovdc(self, rpm):
        """Convert RPM setpoint to VDC
        With safty VDC max = 5.0V et VDCmin = 0.0V
        RPM = VDC*rpmvdcA + rpmvdcB
        or
        RPM = rpmvdcAsup*VDC**2 + rpmvdcBsup*VDC + rpmvdcCsup
        """
        if rpm == 0:
            VDC = 0
        else:
            if ((rpm-self.rpmvdcB)/self.rpmvdcA) <= self.limvdcsup:
                VDC = (rpm-self.rpmvdcB)/self.rpmvdcA
                #print("lin")
            else:
                k = (self.rpmvdcCsup-self.rpmvdcAsup*(self.rpmvdcBsup/(2*self.rpmvdcAsup))**2)
                print(f"not linear part, near pump maximum communication capacity close to {int(k)} rpm\n")
                if rpm > k: #RPM > k
                    print("Warning Maximum pump communication capacity !\n")
                    VDC = -self.rpmvdcBsup/(2*self.rpmvdcAsup)  # VDC = h
                else:
                    VDC = self.polysecdeg_inv(rpm, self.rpmvdcAsup, self.rpmvdcBsup, self.rpmvdcCsup)
            if VDC > 5:
                VDC = 5
            if VDC <0:
                VDC = 0
        return VDC
        
    def tachyread(self, vdc):
        """Convrpmert read vdc to rpm value
        with RPMmax = 200
            self.tachyA = 0.025
            self.tachyB = 0.0066
        VDC = tachyA*RPM + tachyB
        RPM = (VDC-tachyB)/tachyA
        """
        rpm_tachy = (vdc-self.tachyB)/ self.tachyA
        return rpm_tachy
        
    def startcom(self):
        """Srat connection with the labjack
        """
        if self.connection == False:
            self.handle = ljm.openS("T7", "ANY", "ANY")
            self.info = ljm.getHandleInfo(self.handle)
            self.connection = True
            print(f"You play with the {self.info[1]} Labjack T{self.info[0]}")
        else:
            print("Already connected")

    def stopcom(self):
        """Stop connection
        """
        if self.connection == True:
            ljm.eWriteName(self.handle, "DAC0", 0)  #clear output
            ljm.close(self.handle)
            self.connection = False
        else:
            print("Not connected")
    
    def readrpm(self):
        """Give RPM from vdc read in pin AIN0
        """
        if self.connection:
            self.rpm = self.tachyread(ljm.eReadName(self.handle, "AIN0"))
            self.time = datetime.datetime.now()
            return self.rpm, self.time
        else:
            print('Please first used .startcom()')
            
    def update(self):
        """send setpoint to pump
        """
        if self.connection:
            vdc_input = self.rpmtovdc(self.setpoint)
            LJMError = ljm.eWriteName(self.handle, "DAC0", vdc_input)  # put 3V into DAC0 output
            if LJMError:
                print(f"Error n°{LJMError}")
        else:
            print('Please first used .startcom()')
            
    def updatev(self, newsetpoint):
        """send setpoint to pump with spe setpoint
        """
        if self.connection:
            self.setpoint = newsetpoint
            vdc_input = self.rpmtovdc(self.setpoint)
            LJMError = ljm.eWriteName(self.handle, "DAC0", vdc_input)  # put V into DAC0 output
            if LJMError:
                print(f"Error n°{LJMError}")
        else:
            print('Please first used .startcom()')

    def updatetension(self, tension):
        """send setpoint to pump with setpoint in tension
        """
        if self.connection:
            vdc_input = tension
            LJMError = ljm.eWriteName(self.handle, "DAC0", vdc_input)  # put 3V into DAC0 output
            if LJMError:
                print(f"Error n°{LJMError}")
        else:
            print('Please first used .startcom()')
            
    def readtension(self):
        """Give tension in pin AIN0
        """
        if self.connection:
            self.tensionread = ljm.eReadName(self.handle, "AIN0")
            return self.tensionread
        else:
            print('Please first used .startcom()')
            
            
    def changerpmmax(self, rpmmax):
        """Change conv VDC to rpm calib
        self.rpmvdcA = 20.161
        self.rpmvdcB = -0.9726
        """
        if rpmmax == 200:
            self.rpmvdcA = 41.754
            self.rpmvdcB = -50.805
            self.limvdcsup = 3.75
            self.rpmvdcAsup = -27.73
            self.rpmvdcBsup = +253.84
            self.rpmvdcCsup = -456.79
            self.rpmmax = 200
            self.tachyA = 5/rpmmax
            self.tachyB = 0.0065
        elif rpmmax == 100:
            self.rpmvdcA = 20.858
            self.rpmvdcB = -25.366
            self.limvdcsup = 3.75
            self.rpmvdcAsup = -12.641
            self.rpmvdcBsup = 116.66
            self.rpmvdcCsup = -207.07
            self.rpmmax = 100
            self.tachyA = 5/rpmmax
            self.tachyB = 0.0065
        elif rpmmax == 50:
            self.rpmvdcA = 10.441
            self.rpmvdcB = -12.729
            self.limvdcsup = 3.75
            self.rpmvdcAsup = -6.0165
            self.rpmvdcBsup = 56.116
            self.rpmvdcCsup = -99.569
            self.rpmmax = 50
            self.tachyA = 5/rpmmax
            self.tachyB = 0.0065
        elif rpmmax == 25:
            self.rpmvdcA = 5.2188
            self.rpmvdcB = -6.3899
            self.limvdcsup = 3.75
            self.rpmvdcAsup = -3.1334
            self.rpmvdcBsup = 28.877
            self.rpmvdcCsup = -51.131
            self.rpmmax = 25
            self.tachyA = 5/rpmmax
            self.tachyB = 0.0065
        else:
            print(f"{rpmmax}rpm MAX not avaiable\nPlease realise new calibration.\nWith help of .readtension/.readrpm.")



import logging

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.INFO) # logging.WARNING logging.INFO
logger.addHandler(stream_handler)

